
; Pres3D
; (w) 2010 by Volker Schmid (v.schmid@inspirant.de)
;
; Licence: LGPL (http://www.gnu.org/licenses/lgpl.html)
;          english information: http://en.wikipedia.org/wiki/GNU_Lesser_General_Public_License
;          german information:  http://de.wikipedia.org/wiki/LGPL

#Version = "1.09"

CompilerIf #PB_Compiler_OS = #PB_OS_Linux
  #MB_ICONERROR = 0
CompilerEndIf

Enumeration
  #MainWindow
  #GotoWindow
EndEnumeration

Enumeration
  #MenuOpen
  #MenuExit
  #MenuGoto
  #MenuAbout
EndEnumeration

Enumeration
  #PageGadget
  #GoGadget
EndEnumeration

UsePNGImageDecoder()
UsePNGImageEncoder()
UseJPEGImageDecoder()

;IncludeFile "Supersprite3D.pbi" ; this uses the SuperSprite3D include
IncludeFile "Supersprite.pbi" ; this uses the SuperSprite include

; effect settings structure
Structure structEffectParameters
  Type.s
  Direction.s
  Rotation.s
  Misc.s
EndStructure

; objects definition structure
Structure structObject
  Type.i ; 0=char 1=image 2=?
  
  SpriteID.SuperSprite3D ; SuperSprite3D ID
  
  SpriteImageID.i ; ID of the loaded sprite (to free up and copy)
  
  Transform.i ; #True, if transforming is in progress
  
  PosX.f ; desired position X
  PosY.f ; desired position Y
  PosZ.f ; desired position Z
  
  ActualPosX.f ; actual position X
  ActualPosY.f ; actual position Y
  ActualPosZ.f ; actual position Z
  
  AngleX.f ; desired angle X
  AngleY.f ; desired angle Y
  AngleZ.f ; desired angle Z
  
  ActualAngleX.f ; actual angle X
  ActualAngleY.f ; actual angle Y
  ActualAngleZ.f ; actual angle Z
  
  Transparency.l       ; desired Transparency
  ActualTransparency.l ; actual Transparency
  
  Val0.f
  Val1.f
  Val2.f
  Val3.f
  Val4.f ; free variables for effect routines
  Val5.f
  Val6.f
  Val7.f
  Val8.f
  Val9.f
  
  EffectIn.structEffectParameters  ; effect ID for ingoing
  EffectOut.structEffectParameters ; effect ID for outgoing
  
  GroupID.i    ; ID in my group (1 to GroupMax.l)
  GroupMax.i   ; max ID in my group
  
  Delay.i      ; delay in milliseconds
  
  ImageSplitSize.i ; split image into tiles with this size
  
EndStructure

; program settings structure
Structure structSettings
  ScreenFull.i    ; #True or #False
  ScreenWidth.i
  ScreenHeight.i
  BackgroundColor.l
  Smoothing.i     ; #True or #False
EndStructure

#BufferImages = 6000 ; maximum number of elements per page
#XML          = 0    ; XML-ID
#InitialZ     = 578  ; initial Z-Position of (currently) all new elements

Global *XML_CurrentPageNode         ; holds the next XML node to read the next page from
Global XML_FilePath.s               ; path (incl. last \) of the currently loaded xml file
Global XML_CurrentXMLFile.s         ; the currently loaded xml file
Global XML_Settings.structSettings  ; global settings of the currently loaded XML
Global *XML_MainNode                ; global XML main node

Global IsFullscreen.i               ; #True if Fullscreen-Mode is active, #False if not

Global Dim InObjects.structObject(#BufferImages)
Global InObjectCount.i = 0 ; counter for ingoing objects
Global Dim OutObjects.structObject(#BufferImages)
Global OutObjectCount.i = 0 ; counter for outgoing objects

Structure NodeAddress
  *Node
EndStructure

Global Dim PageNo.NodeAddress(0)

CompilerIf #PB_Compiler_OS = #PB_OS_Windows
  ; allow Sprite3D view from behind (windows-only!!!!!!)
  Define pd3d.IDirect3DDevice9
CompilerEndIf

InitSprite()
;InitSuperSprite3D()
InitSuperSprite()
InitKeyboard()
InitMouse()

; ---------------------------------------------------------------
; Effects
; ---------------------------------------------------------------
IncludeFile "Effects.pbi"

; ---------------------------------------------------------------
; Procedures
; ---------------------------------------------------------------
  
; open a screen for displaying
Procedure OpenDisplayScreen()
  
  If IsWindow(0) <> 0
    ; close previous opened window
    CloseScreen()
    CloseWindow(0)
  EndIf
  
  ScreenRet = OpenScreen(XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, 32, "Pres3D")
  
  IsFullscreen = #True
  
  ReleaseMouse(0)
  
  If XML_Settings\Smoothing = #True
    ;Sprite3DQuality(#PB_Sprite3D_BilinearFiltering)
    SpriteQuality(#PB_Sprite_BilinearFiltering)
  Else
    ;Sprite3DQuality(#PB_Sprite3D_NoFiltering)
    SpriteQuality(#PB_Sprite_NoFiltering)
  EndIf
  
  ProcedureReturn ScreenRet
  
EndProcedure

; open a window for displaying
Procedure OpenDisplayWindow()
  
  If IsFullscreen : CloseScreen() : EndIf
  
  If Not OpenWindow(#MainWindow, 0, 0, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight + 43, "Pres3D V" + #Version, #PB_Window_ScreenCentered | #PB_Window_SystemMenu)
    MessageRequester("error", "Can not open window! Please check your values...", #MB_ICONERROR)
    End
  EndIf
  
  CreateMenu(0, WindowID(0))
  MenuTitle("File")
  MenuItem(#MenuOpen, "Open")
  MenuBar()
  MenuItem(#MenuExit, "Exit")
  MenuTitle("Presentation")
  MenuItem(#MenuGoto, "Goto page")
  MenuTitle("Help")
  MenuItem(#MenuAbout, "About")
  
  If CreateStatusBar(0, WindowID(0))
    AddStatusBarField(80)
    AddStatusBarField(80)
  EndIf
  
  ScreenRet = OpenWindowedScreen(WindowID(#MainWindow), 0, 0, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, 0, 0, 0)
  
  IsFullscreen = #False
  
  ReleaseMouse(1)
  
  If XML_Settings\Smoothing = #True
    ;Sprite3DQuality(#PB_Sprite3D_BilinearFiltering)
    SpriteQuality(#PB_Sprite_BilinearFiltering)
  Else
    ;Sprite3DQuality(#PB_Sprite3D_NoFiltering)
    SpriteQuality(#PB_Sprite_NoFiltering)
  EndIf

  ProcedureReturn ScreenRet
  
EndProcedure

; ImageToSprite function copy's an image to a sprite without the need to safe the
; image as png before. Written by STARGTE (german PureBasic Forum).
; Adapted and reformated by V. Schmid (Oct 2010).
Procedure ImageToSprite(Sprite.i, Image.i, Mode.i = #False)
  Protected Row.i
  Protected ImageHeight.i = ImageHeight(Image)
  Protected ImageWidth.i  = ImageWidth(Image)
  Protected *ReadBuffer,  ReadPitch.i,  ReadFormat.i
  Protected *WriteBuffer, WritePitch.i, WriteFormat.i
  StartDrawing(ImageOutput(Image))
    *ReadBuffer = DrawingBuffer()
    ReadPitch   = DrawingBufferPitch()
    ReadFormat  = DrawingBufferPixelFormat()
  StopDrawing()
  If Sprite = #PB_Any
    Sprite = CreateSprite(#PB_Any, ImageWidth, ImageHeight, Mode)
  Else
    
    CreateSprite(Sprite, ImageWidth, ImageHeight, Mode)
  EndIf
  StartDrawing(SpriteOutput(Sprite))
    *WriteBuffer = DrawingBuffer()
    WritePitch   = DrawingBufferPitch()
    WriteFormat  = DrawingBufferPixelFormat()
    
    If WriteFormat & #PB_PixelFormat_8Bits : PixelSize = 1 : EndIf
    If WriteFormat & #PB_PixelFormat_15Bits  Or WriteFormat & #PB_PixelFormat_16Bits : PixelSize = 2 : EndIf
    If WriteFormat & #PB_PixelFormat_24Bits_RGB  Or WriteFormat & #PB_PixelFormat_24Bits_BGR : PixelSize = 3 : EndIf
    If WriteFormat & #PB_PixelFormat_32Bits_RGB  Or WriteFormat & #PB_PixelFormat_32Bits_BGR : PixelSize = 4 : EndIf
    
    If ReadFormat & #PB_PixelFormat_ReversedY XOr WriteFormat & #PB_PixelFormat_ReversedY
      For Row = 0 To ImageHeight - 1
        CopyMemory(*ReadBuffer + (ImageHeight - 1 - Row) * ReadPitch, *WriteBuffer + Row * WritePitch, ImageWidth * PixelSize)
      Next
    Else
      For Row = 0 To ImageHeight - 1
        CopyMemory(*ReadBuffer + Row * ReadPitch, *WriteBuffer + Row * WritePitch, ImageWidth * PixelSize)
      Next
    EndIf
  StopDrawing()
  ProcedureReturn Sprite
EndProcedure

; Returns the width (pixel) of a char respecting it's surrounding chars...
; IMPORTANT: This needs a current StartDrawing() context and a defined font!
Procedure.i GetCharWidth(Text.s, CharPos.i)
  If CharPos < 2
    ; first char
    Char1$ = Left(Text, 2) ; the first two
    Char2$ = Mid(Text, 2, 1); without the needed
    ProcedureReturn TextWidth(Char1$) - TextWidth(Char2$)
  EndIf
  If CharPos > Len(Text) - 1
    ; last char
    Char1$ = Right(Text, 2) ; the last two
    Char2$ = Right(Text, 1); without the needed
    ProcedureReturn TextWidth(Char1$) - TextWidth(Char2$)
  EndIf
  If CharPos > 1 And CharPos < Len(Text)
    ; all other chars
    Char1$ = Mid(Text, CharPos - 1, 3) ; all three
    Char2$ = Left(Char1$, 1) + Right(Char1$, 1); without the needed
    ProcedureReturn TextWidth(Char1$) - TextWidth(Char2$)
  EndIf
EndProcedure

; return a free SpriteID that is currently and definitely not used as sprite
Procedure.i GetFreeSpriteID()
  Protected SpriteID.i

  Repeat : SpriteID = Random(50000) : Until Not IsSprite(SpriteID)
  ProcedureReturn SpriteID
EndProcedure

; insert a new group ob objects containing text to the ingoing objects
Procedure MakeTextGroup(Text.s, FontName.s, Size.f, PosX.i, PosY.i, Color.i, Style.i, Delay.l, *EffectIn.structEffectParameters, *EffectOut.structEffectParameters)
  
  LoadFont(0, FontName, Size, Style) ; load Arial font in desired size
  
  BoxSize = 16
  
  If Size > 15.0 : BoxSize = 32 : EndIf
  If Size > 20.0 : BoxSize = 64 : EndIf
  If Size > 40.0 : BoxSize = 128 : EndIf
  
  ColR = Red(Color)
  ColG = Green(Color)
  ColB = Blue(Color)
  
  ; generate char sprites
  
  LeftPos = 0; init
  For x = 1 To Len(Text)
    
    InObjectCount + 1
    
    ; create sprite
    CreateImage(0, BoxSize, BoxSize, 32, #PB_Image_Transparent)
    StartDrawing(ImageOutput(0))
    Box(0, 0, BoxSize, BoxSize, RGBA(255,255,255,0))
    
    ; set font
    DrawingFont(FontID(0))
    
    MaxWi = TextWidth("_")
    MaxHe = TextHeight("|")
    
    CenterX = - TextWidth(Text.s) / 2
    CenterY = - TextHeight(Text.s) / 2
    
    Char.s = Mid(Text, x, 1)
    Wi = GetCharWidth(Text, x)
    
    ; create object structure
    With InObjects(InObjectCount)
      \Type         = 0; text
      \PosX         = PosX + LeftPos + CenterX + MaxWi / 2
      \PosY         = PosY; + CenterY
      \PosZ         = #InitialZ
      \ActualPosX   = \PosX
      \ActualPosY   = \PosY
      \ActualPosZ   = \PosZ
      
      \Transparency        = 255
      \ActualTransparency  = \Transparency
      
      \EffectIn\Type       = *EffectIn\Type
      \EffectIn\Direction  = *EffectIn\Direction
      \EffectIn\Rotation   = *EffectIn\Rotation
      \EffectIn\Misc       = *EffectIn\Misc
      
      \EffectOut\Type      = *EffectOut\Type
      \EffectOut\Direction = *EffectOut\Direction
      \EffectOut\Rotation  = *EffectOut\Rotation
      \EffectOut\Misc      = *EffectOut\Misc
      
      \GroupID      = x
      \GroupMax     = Len(Text)
      
      \Transform    = #True
      
      ; init
      \ActualAngleX = 0
      \ActualAngleY = 0
      \ActualAngleZ = 0
      \AngleX       = 0
      \AngleY       = 0
      \AngleZ       = 0
      
      \Delay        = Delay
      
    EndWith
    LeftPos + Wi
    
    ; draw char to image
    DrawingMode(#PB_2DDrawing_AllChannels)
    ; DrawingMode(#PB_2DDrawing_AllChannels | #PB_2DDrawing_Gradient)
    ; CircularGradient(BoxSize.l/2, BoxSize.l/2, BoxSize.l/3) : GradientColor(0.0,RGBA(0,255,0,255)) : GradientColor(1.0,RGBA(0,0,0, 255))
    DrawText((BoxSize - MaxWi) / 2, (BoxSize - MaxHe) / 2, Char, RGBA(ColR, ColG, ColB, 255), RGBA(0,0,0,0))
    StopDrawing()
    
    ; load ingoing image as sprite
    SpriteID = GetFreeSpriteID()
    ;ImageToSprite(SpriteID, 0, #PB_Sprite_Texture|#PB_Sprite_AlphaBlending)
    ImageToSprite(SpriteID, 0, #PB_Sprite_AlphaBlending)
    InObjects(InObjectCount)\SpriteImageID = SpriteID
    
    ;CreateSuperSprite3D(InObjects(InObjectCount)\SpriteID, SpriteID, BoxSize, BoxSize)
    CreateSuperSprite(InObjects(InObjectCount)\SpriteID, SpriteID, BoxSize, BoxSize)
    
    EffectInitIn(*EffectIn, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, InObjectCount, x, Len(Text))
    
    FreeImage(0)
    
  Next
  
EndProcedure

; insert a new group ob objects containing text to the ingoing objects
Procedure MakeTextBlock(Text.s, FontName.s, Size.f, PosX.i, PosY.i, Color.i, Style.i, Delay.i, *EffectIn.structEffectParameters, *EffectOut.structEffectParameters)
  
  LoadFont(0, FontName, Size, Style) ; load Arial font in desired size
  
  InObjectCount + 1
  
  ; create test-image (for getting the needed width)
  CreateImage(0, 512, 512, 32, #PB_Image_Transparent)
  StartDrawing(ImageOutput(0))
  ; set font
  DrawingFont(FontID(0))
  ; get needed size
  Size.f = TextWidth(Text)
  ; free it
  StopDrawing()
  FreeImage(0)
  
  BoxSize = 16
  If Size > 16.0 : BoxSize = 32: EndIf
  If Size > 32.0 : BoxSize = 64: EndIf
  If Size > 64.0 : BoxSize = 128: EndIf
  If Size > 128.0 : BoxSize = 256: EndIf
  If Size > 256.0 : BoxSize = 512: EndIf
  If Size > 512.0 : BoxSize = Round(Size / 10.0, #PB_Round_Up) * 10 : EndIf
  
  ; create object structure
  With InObjects(InObjectCount)
    \Type         = 0; text
    \PosX         = PosX; - Size.f / 2
    \PosY         = PosY; - Size.f / 2
    \PosZ         = #InitialZ
    \ActualPosX   = \PosX
    \ActualPosY   = \PosY
    \ActualPosZ   = \PosZ
    
    \Transparency        = 255
    \ActualTransparency  = \Transparency
    
    \EffectIn\Type       = *EffectIn\Type
    \EffectIn\Direction  = *EffectIn\Direction
    \EffectIn\Rotation   = *EffectIn\Rotation
    \EffectIn\Misc       = *EffectIn\Misc
    
    \EffectOut\Type      = *EffectOut\Type
    \EffectOut\Direction = *EffectOut\Direction
    \EffectOut\Rotation  = *EffectOut\Rotation
    \EffectOut\Misc      = *EffectOut\Misc
    
    \GroupID      = 1
    \GroupMax     = 1
    
    \Transform    = #True
    
    ; init
    \ActualAngleX = 0
    \ActualAngleY = 0
    \ActualAngleZ = 0
    \AngleX       = 0
    \AngleY       = 0
    \AngleZ       = 0
    
    \Delay        = Delay
    
  EndWith
  
  ColR = Red(Color)
  ColG = Green(Color)
  ColB = Blue(Color)
  
  ; create sprite
  CreateImage(0, BoxSize, BoxSize, 32, #PB_Image_Transparent)
  StartDrawing(ImageOutput(0))
  Box(0, 0, BoxSize, BoxSize, RGBA(255,255,255,0))
  
  ; set font
  DrawingFont(FontID(0))
  
  CenterX = (BoxSize - TextWidth(Text)) / 2
  CenterY = (BoxSize - TextHeight(Text)) / 2
    
  ; draw char to image
  DrawingMode(#PB_2DDrawing_AllChannels)
  ; Box(0, 0, BoxSize.l, BoxSize.l, RGBA(55,55,75,255))
  DrawText(CenterX, CenterY, Text, RGBA(ColR, ColG, ColB, 255), RGBA(0,0,0,0))
  StopDrawing()
  
  ; load ingoing image as sprite
  SpriteID = GetFreeSpriteID()
  ;ImageToSprite(SpriteID, 0, #PB_Sprite_Texture|#PB_Sprite_AlphaBlending)
  ImageToSprite(SpriteID, 0, #PB_Sprite_AlphaBlending)
  InObjects(InObjectCount)\SpriteImageID = SpriteID
  
  ;CreateSuperSprite3D(InObjects(InObjectCount)\SpriteID, SpriteID, BoxSize, BoxSize)
  CreateSuperSprite(InObjects(InObjectCount)\SpriteID, SpriteID, BoxSize, BoxSize)
  
  EffectInitIn(*EffectIn, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, InObjectCount, 1, 1)
  
  FreeImage(0)
EndProcedure

; insert a new image to the ingoing objects
Procedure.i MakeImage(Filename.s, PosX.i, PosY.i, Width.i, Height.i, Delay.i, *EffectIn.structEffectParameters, *EffectOut.structEffectParameters)
  
  If LoadImage(0, Filename) = 0
    ProcedureReturn 0
  EndIf
  
  InObjectCount + 1
    
  
  If Width < 1 : Width  = ImageWidth(0) : EndIf
  If Height < 1: Height = ImageHeight(0): EndIf
  
  If ImageWidth(0) > ImageHeight(0)
    ; landscape
    Widest = Width
  Else
    ; portrait
    Widest = Height
  EndIf
  For x = 4 To 18
    If Pow(2, x) >= Widest
      BoxSize = Pow(2, x)
      Break
    EndIf
  Next
  
  ; create object structure
  With InObjects(InObjectCount)
    \Type         = 1; image
    \PosX         = PosX ;- (BoxSize.l / 2)
    \PosY         = PosY ;- (BoxSize.l / 2)
    \PosZ         = #InitialZ
    \ActualPosX   = \PosX
    \ActualPosY   = \PosY
    \ActualPosZ   = \PosZ
    
    \Transparency        = 255
    \ActualTransparency  = \Transparency
    
    \EffectIn\Type       = *EffectIn\Type
    \EffectIn\Direction  = *EffectIn\Direction
    \EffectIn\Rotation   = *EffectIn\Rotation
    \EffectIn\Misc       = *EffectIn\Misc
    
    \EffectOut\Type      = *EffectOut\Type
    \EffectOut\Direction = *EffectOut\Direction
    \EffectOut\Rotation  = *EffectOut\Rotation
    \EffectOut\Misc      = *EffectOut\Misc
    
    \Transform    = #True
    
    ; init
    \ActualAngleX = 0
    \ActualAngleY = 0
    \ActualAngleZ = 0
    \AngleX       = 0
    \AngleY       = 0
    \AngleZ       = 0
    
    \GroupID      = 1
    \GroupMax     = 1
    
    \Delay        = Delay
    
  EndWith
  
  CreateImage(1, BoxSize, BoxSize, 32,  #PB_Image_Transparent)
  
  StartDrawing(ImageOutput(1))
    DrawingMode(#PB_2DDrawing_AlphaBlend)
    ResizeImage(0, Width, Height, #PB_Image_Smooth)
    DrawImage(ImageID(0), (BoxSize - Width) / 2, (BoxSize - Height) / 2, Width, Height)
    ; Box(0,0,BoxSize.l, BoxSize.l, RGBA(200,200,200,200)) ; for debug only
  StopDrawing()
  
  SpriteID = GetFreeSpriteID()
  ;ImageToSprite(SpriteID, 1, #PB_Sprite_Texture|#PB_Sprite_AlphaBlending)
  ImageToSprite(SpriteID, 1, #PB_Sprite_AlphaBlending)
  InObjects(InObjectCount)\SpriteImageID = SpriteID
  
  ;CreateSuperSprite3D(InObjects(InObjectCount)\SpriteID, SpriteID, SpriteWidth(SpriteID), SpriteHeight(SpriteID))
  CreateSuperSprite(InObjects(InObjectCount)\SpriteID, SpriteID, SpriteWidth(SpriteID), SpriteHeight(SpriteID))
  
  EffectInitIn(*EffectIn, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, InObjectCount, 1, 1)
  
  FreeImage(1)
  FreeImage(0)
  
EndProcedure

; insert a new image to the ingoing objects. This one is split into multiple smaller quadratic images.
Procedure.i MakeImageSplit(Filename.s, PosX.i, PosY.i, Width.i, Height.i, Delay.i, SplitSize.i, *EffectIn.structEffectParameters, *EffectOut.structEffectParameters)
  
  If LoadImage(0, Filename) = 0 ; load original image
    ProcedureReturn 0
  EndIf
  
  If Width < 1 : Width = ImageWidth(0) : EndIf
  If Height < 1: Height = ImageHeight(0): EndIf
  
  If ImageWidth(0) > ImageHeight(0)
    ; landscape
    Widest = Width
  Else
    ; portrait
    Widest = Height
  EndIf
  For x = 4 To 18
    If Pow(2, x) >= Widest
      BoxSize = Pow(2, x)
      Break
    EndIf
  Next
  
  ResizeImage(0, Width, Height, #PB_Image_Smooth)
  
  CreateImage(1, BoxSize, BoxSize, 32, #PB_Image_Transparent) ; create scaled image
  
  StartDrawing(ImageOutput(1)); start drawing to scaled image
  DrawingMode(#PB_2DDrawing_AlphaBlend)
  ; Box(1,1,BoxSize.l, BoxSize.l, RGBA(100,100,100,100))
  DrawImage(ImageID(0), (BoxSize - Width) / 2, (BoxSize - Height) / 2, Width, Height); insert original image to this block
  StopDrawing()
  
  ; in image 1 is the correct and complete image to display
  
  Parts = BoxSize / SplitSize
  
  For x = 0 To Parts - 1
    For y = 0 To Parts - 1
      
      InObjectCount + 1
      
      ; create object structure
      With InObjects(InObjectCount)
        \Type         = 1; image
        \PosX         = PosX + (x * SplitSize) - (BoxSize / 2) + Splitsize / 2
        \PosY         = PosY + (y * SplitSize) - (BoxSize / 2) + SplitSize / 2
        \PosZ         = #InitialZ
        \ActualPosX   = \PosX
        \ActualPosY   = \PosY
        \ActualPosZ   = \PosZ
        
        \Transparency        = 255
        \ActualTransparency  = \Transparency
        
        \EffectIn\Type       = *EffectIn\Type
        \EffectIn\Direction  = *EffectIn\Direction
        \EffectIn\Rotation   = *EffectIn\Rotation
        \EffectIn\Misc       = *EffectIn\Misc
        
        \EffectOut\Type      = *EffectOut\Type
        \EffectOut\Direction = *EffectOut\Direction
        \EffectOut\Rotation  = *EffectOut\Rotation
        \EffectOut\Misc      = *EffectOut\Misc
        
        \Transform    = #True
        
        ; init
        \ActualAngleX = 0
        \ActualAngleY = 0
        \ActualAngleZ = 0
        \AngleX       = 0
        \AngleY       = 0
        \AngleZ       = 0
        
        \GroupID      = y * 3
        \GroupMax     = Parts * 3
        
        \Delay        = Delay
        
      EndWith
      
      StartDrawing(ImageOutput(1)); start drawing to scaled image
      
      GrabDrawingImage(2, x * SplitSize, y * SplitSize, SplitSize, SplitSize)
      
      StopDrawing()
      
      SpriteID = GetFreeSpriteID()
      ;ImageToSprite(SpriteID, 2, #PB_Sprite_Texture|#PB_Sprite_AlphaBlending)
      ImageToSprite(SpriteID, 2, #PB_Sprite_AlphaBlending)
      
      InObjects(InObjectCount)\SpriteImageID = SpriteID
      
      ;CreateSuperSprite3D(InObjects(InObjectCount)\SpriteID, SpriteID, SpriteWidth(SpriteID), SpriteHeight(SpriteID))
      CreateSuperSprite(InObjects(InObjectCount)\SpriteID, SpriteID, SpriteWidth(SpriteID), SpriteHeight(SpriteID))
      
      EffectInitIn(*EffectIn, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, InObjectCount, y * 3, Parts * 3)
      
      FreeImage(2)
      
    Next y
  Next x
  
  FreeImage(1)  ; free scaled image
  
  FreeImage(0)  ; free initial image
  
EndProcedure

; Make all current ingoing objects to outgoing objects (will start to fade out)
Procedure ActivateOutgoing()
  
  ; clean all current outgoing objects (their animation has been finished)
  For x = 1 To OutObjectCount
    If OutObjects(x)\PosX > 0
      ;FreeSprite3D(OutObjects(x)\SpriteImageID)
      FreeSprite(OutObjects(x)\SpriteImageID)
      ; not shure, if it needs some kind of killing the assigned SuperSprite3D instance here?
    EndIf
  Next
  
  Dim OutObjects.structObject(#BufferImages) ; clean array
  
  ; copy all currently displayed objects (InObjects) to outgoing
  For x = 1 To InObjectCount
    ; copy
    OutObjects.structObject(x) = InObjects.structObject(x)
    
    ; start their outgoing effect
    *Out.structEffectParameters = OutObjects(x)\EffectOut
    EffectInitOut(*Out, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, x, OutObjects(x)\GroupID, OutObjects(x)\GroupMax)
  Next x
  
  OutObjectCount = InObjectCount
  
  ; clean old ingoing objects
  Dim InObjects.structObject(#BufferImages) ; clean array for ingoing objects
  InObjectCount = 0
EndProcedure

; ---------------------------------------------------------------
; XML handling
; ---------------------------------------------------------------

; get the effect (effectin or effectout) parameters in the given XML node
Procedure.l GetEffectNode(*XMLNode, EffectName.s)
  Protected *Result.structEffectParameters = AllocateMemory(SizeOf(structEffectParameters))
  
  EffectName = LCase(EffectName)
  
  Count = XMLChildCount(*XMLNode)
  
  For x = 1 To Count
    *ChildNode = ChildXMLNode(*XMLNode, x)
    Name.s = LCase(GetXMLNodeName(*ChildNode))
    If Name = EffectName
      ; effect found
      *Result\Type      = GetXMLAttribute(*ChildNode, "type")
      *Result\Direction = GetXMLAttribute(*ChildNode, "direction")
      *Result\Rotation  = GetXMLAttribute(*ChildNode, "rotation")
      *Result\Misc      = GetXMLAttribute(*ChildNode, "misc")
      ProcedureReturn *Result
    EndIf
  Next
  
EndProcedure

; get all text and images from the current page (*XMLNode) to the ingoing objects
Procedure InsertXMLElements(*XMLNode)
  *ChildNode = ChildXMLNode(*XMLNode)
  While *ChildNode <> 0
    If *ChildNode <> 0
      ; node found
      NodeName.s = GetXMLNodeName(*ChildNode)
      ; Debug "Node: " + NodeName
      
      Select NodeName
        Case "text"
          Caption.s   =     GetXMLAttribute(*ChildNode, "caption")
          PosX.i      = Val(GetXMLAttribute(*ChildNode, "posx"))
          PosY.i      = Val(GetXMLAttribute(*ChildNode, "posy"))
          FontName.s  =     GetXMLAttribute(*ChildNode, "fontname")
          FontSize.f  = Val(GetXMLAttribute(*ChildNode, "fontsize"))
          FontStyle.i = Val(GetXMLAttribute(*ChildNode, "fontstyle"))
          Color.s     =     GetXMLAttribute(*ChildNode, "color")
          Delay.i     = Val(GetXMLAttribute(*ChildNode, "delay"))
          
          *EffectIn.structEffectParameters  = GetEffectNode(*ChildNode, "effectin")
          *EffectOut.structEffectParameters = GetEffectNode(*ChildNode, "effectout")
          
          Color       = ReplaceString(Color, "#" , ""); ignore # (html style)
          
          ColorVal.i  = RGB(Val("$"+Mid(Color,1,2)),Val("$"+Mid(Color,3,2)),Val("$"+Mid(Color,5,2))); parse hex color values
          
          MakeTextGroup(Caption, FontName, FontSize, PosX, PosY, ColorVal, FontStyle, Delay, *EffectIn, *EffectOut)
          
        Case "textblock"
          Caption.s   =     GetXMLAttribute(*ChildNode, "caption")
          PosX.i      = Val(GetXMLAttribute(*ChildNode, "posx"))
          PosY.i      = Val(GetXMLAttribute(*ChildNode, "posy"))
          FontName.s  =     GetXMLAttribute(*ChildNode, "fontname")
          FontSize.f  = Val(GetXMLAttribute(*ChildNode, "fontsize"))
          FontStyle.i = Val(GetXMLAttribute(*ChildNode, "fontstyle"))
          Color.s     =     GetXMLAttribute(*ChildNode, "color")
          Delay.i     = Val(GetXMLAttribute(*ChildNode, "delay"))
          
          *EffectIn.structEffectParameters  = GetEffectNode(*ChildNode, "effectin")
          *EffectOut.structEffectParameters = GetEffectNode(*ChildNode, "effectout")
          
          Color       = ReplaceString(Color, "#" , ""); ignore # (html style)
          
          ColorVal.i  = RGB(Val("$"+Mid(Color,1,2)),Val("$"+Mid(Color,3,2)),Val("$"+Mid(Color,5,2))); parse hex color values
          
          MakeTextBlock(Caption, FontName, FontSize, PosX, PosY, ColorVal, FontStyle, Delay, *EffectIn, *EffectOut)
          
        Case "image"
          PosX.i      = Val(GetXMLAttribute(*ChildNode, "posx"))
          PosY.i      = Val(GetXMLAttribute(*ChildNode, "posy"))
          Width.i     = Val(GetXMLAttribute(*ChildNode, "width"))
          Height.i    = Val(GetXMLAttribute(*ChildNode, "height"))
          Image.s     =     GetXMLAttribute(*ChildNode, "image")
          Delay.i     = Val(GetXMLAttribute(*ChildNode, "delay"))
          SplitSize.i = Val(GetXMLAttribute(*ChildNode, "splitsize"))
          
          *EffectIn.structEffectParameters  = GetEffectNode(*ChildNode, "effectin")
          *EffectOut.structEffectParameters = GetEffectNode(*ChildNode, "effectout")
          
          If SplitSize >= 4
            MakeImageSplit(Image, PosX, PosY, Width, Height, Delay, SplitSize, *EffectIn, *EffectOut)
          Else
            MakeImage(Image, PosX, PosY, Width, Height, Delay, *EffectIn, *EffectOut)
          EndIf
          
      EndSelect
      
    EndIf
    
    *ChildNode = NextXMLNode(*ChildNode)
  Wend
  
EndProcedure

; remove all objects
Procedure ClearObjects()
  InObjectCount = 0
  For i = 0 To ArraySize(InObjects())
;     If IsSprite3D(InObjects(i)\SpriteID)
;       FreeSprite3D(InObjects(i)\SpriteID)
;     EndIf
    If IsSprite(InObjects(i)\SpriteImageID)
      FreeSprite(InObjects(i)\SpriteImageID)
    EndIf
  Next i
  
  ReDim InObjects.structObject(0)
  ReDim InObjects.structObject(#BufferImages)
  
  OutObjectCount = 0
  For i = 0 To ArraySize(OutObjects())
;     If IsSprite3D(OutObjects(i)\SpriteID)
;       FreeSprite3D(OutObjects(i)\SpriteID)
;     EndIf
    If IsSprite(OutObjects(i)\SpriteImageID)
      FreeSprite(OutObjects(i)\SpriteImageID)
    EndIf
  Next i
  
  ReDim OutObjects.structObject(0)
  ReDim OutObjects.structObject(#BufferImages)  
EndProcedure

; open the XMLFile.s for a slideshow
Procedure.i OpenSlideXML(XMLFile.s)
  
  If IsXML(#XML) <> 0
    ; cleanup current presentation
    FreeXML(#XML)
    *XML_CurrentPageNode   = #Null
    XML_CurrentXMLFile.s  = ""
    XML_FilePath.s        = ""
    
    ClearObjects()
    
  EndIf
  
  If LoadXML(#XML, XMLFile) = 0
    XMLFile = OpenFileRequester("Choose a Presentation", GetPathPart(ProgramFilename()) + XMLFile, "XML (*.xml)|*.xml", 0 )
    If LoadXML(#XML, XMLFile) = 0
      ProcedureReturn #False
    EndIf
  EndIf
 
  Repeat
    If XMLStatus(#XML) <> #PB_XML_Success
      res = MessageRequester("Info", "Sorry, this was no suitable slide-file. Like to open another one ?", #PB_MessageRequester_YesNo| $50)
      If res = #PB_MessageRequester_Yes       
        XMLFile = OpenFileRequester("Again", GetPathPart(ProgramFilename()), "XML (*.xml)|*.xml", 0 )
        If LoadXML(#XML, XMLFile) = 0
          ProcedureReturn #False
        EndIf
      Else
        MessageRequester("Oh Well", "so maybe some other time ", $50)
        ProcedureReturn #False
      EndIf
    EndIf
  Until XMLStatus(#XML) = #PB_XML_Success
  
  
  If XMLStatus(#XML) <> #PB_XML_Success
    Message.s = "Error in the XML file: "
    Message + "Message: " + XMLError(#XML) + " "
    Message + "Line: " + Str(XMLErrorLine(#XML)) + " Character: " + Str(XMLErrorPosition(#XML))
    Debug Message
    ProcedureReturn #False
  EndIf
  
  *XML_MainNode = MainXMLNode(#XML)
  
  XMLName.s = GetXMLNodeName(*XML_MainNode)
  If XMLName <> "pres3d"
    Debug "No valid Pres3D xml file!"
    ProcedureReturn #False
  EndIf
  
  *XML_CurrentPageNode = ChildXMLNode(*XML_MainNode)
  XML_FilePath          = GetPathPart(XMLFile)
  XML_CurrentXMLFile.s  = XMLFile
  
  ; --- read settings in current xml structure ---
  
  *ChildNode = ChildXMLNode(*XML_MainNode)
  ; defaults
  XML_Settings\ScreenFull       = 0
  XML_Settings\ScreenWidth      = 800
  XML_Settings\ScreenHeight     = 600
  XML_Settings\BackgroundColor  = RGB(255,255,255)
  ; read values (if given)
  While *ChildNode <> 0
    If *ChildNode <> 0 And XMLNodeType(*ChildNode) = #PB_XML_Normal 
      ; node found
      NodeName.s = GetXMLNodeName(*ChildNode)
      If NodeName.s = "settings"
        XML_Settings\ScreenFull   = Val(GetXMLAttribute(*ChildNode, "screenfull"))
        XML_Settings\ScreenWidth  = Val(GetXMLAttribute(*ChildNode, "screenwidth"))
        XML_Settings\ScreenHeight = Val(GetXMLAttribute(*ChildNode, "screenheight"))
        Color.s                   =     GetXMLAttribute(*ChildNode, "backgroundcolor")
        If GetXMLAttribute(*ChildNode, "smoothing") = "true" Or GetXMLAttribute(*ChildNode, "smoothing") = "1"
          XML_Settings\Smoothing  = #True
        Else
          XML_Settings\Smoothing  = #False
        EndIf
        
        Color = ReplaceString(Color, "#" , ""); ignore # (html style)  
        ColorVal.i = RGB(Val("$"+Mid(Color,1,2)),Val("$"+Mid(Color,3,2)),Val("$"+Mid(Color,5,2))); parse hex color values
        XML_Settings\BackgroundColor = ColorVal
        ; Debug "Settings loaded..."
      EndIf
      
    EndIf
    
    *ChildNode = NextXMLNode(*ChildNode)
  Wend
  
  If XML_Settings\ScreenWidth = 0 Or XML_Settings\ScreenHeight = 0
    ; use desktop settings
    ExamineDesktops()
    XML_Settings\ScreenWidth = DesktopWidth(0)
    XML_Settings\ScreenHeight = DesktopHeight(0)
  Else
    
    If ExamineScreenModes()
      While NextScreenMode()
        If XML_Settings\ScreenWidth <= ScreenModeWidth() And XML_Settings\ScreenHeight <= ScreenModeHeight()
          XML_Settings\ScreenWidth = ScreenModeWidth()
          XML_Settings\ScreenHeight = ScreenModeHeight()
          Break
        EndIf
      Wend
    EndIf
    
  EndIf
  
;  Debug XML_Settings\ScreenWidth
;  Debug XML_Settings\ScreenHeight
  
  *XML_CurrentPageNode = ChildXMLNode(*XML_MainNode)
  
  PageNo = 0
  Repeat
    *XML_CurrentPageNode = NextXMLNode(*XML_CurrentPageNode)
    If *XML_CurrentPageNode
      If XMLNodeType(*XML_CurrentPageNode) = #PB_XML_Normal
        If GetXMLNodeName(*XML_CurrentPageNode) = "page"
          PageNo + 1
          ReDim PageNo(PageNo)
          PageNo(PageNo)\Node = *XML_CurrentPageNode
        EndIf  
      EndIf
    EndIf
  Until Not *XML_CurrentPageNode
  
  ProcedureReturn #True
  
EndProcedure

; jump to an absolute page
Procedure.i GotoXMLPage(No)
  If No < 0 Or No > ArraySize(PageNo())
    ProcedureReturn #False
  Else
    *XML_CurrentPageNode = PageNo(No)\Node
    InsertXMLElements(*XML_CurrentPageNode)
    If IsStatusBar(0) : StatusBarText(0, 0, "Page " + Str(No) + " of " + Str(ArraySize(PageNo())), #PB_StatusBar_Center) : EndIf
  EndIf
  
  ProcedureReturn #True
  
EndProcedure



; ---------------------------------------------------------------
; Init
; ---------------------------------------------------------------

If CountProgramParameters()
  Filename$ = ProgramParameter(0)
Else
  Filename$ = "slide.xml"
EndIf

; load xml definition
If OpenSlideXML(Filename$) = #False
  Debug "Can not load XML file"
;  End
EndIf

; ---------------------------------------------------------------
; Screen
; ---------------------------------------------------------------
If XML_Settings\ScreenFull = 0
  ScreenRet = OpenDisplayWindow(); windowed
Else
  ScreenRet = OpenDisplayScreen(); fullscreen-mode
EndIf

If ScreenRet = 0
  ; screen successfully opened?
  MessageRequester("error", "Can not open screen! Please check your XML settings values...", #MB_ICONERROR)
  End
EndIf

SetFrameRate(60); max 60 frames

; load first page
;If LoadNextXMLPage() = #True
PageCounter = 1
If Not GotoXMLPage(PageCounter)
  Debug "No valid page information found in XML definition file!"
  End
EndIf

; define the center of the screen (perspective)
SetPerspective(XML_Settings\ScreenWidth / 2, XML_Settings\ScreenHeight / 2)

; ---------------------------------------------------------------
; Main Loop
; ---------------------------------------------------------------
PageStartTime = ElapsedMilliseconds()
;LastMousePush = ElapsedMilliseconds()
Quit = #False

PageActive = #False
Zoom = 0
NewZoom = 0
WM_wheel.w = 0

ActualWidth = XML_Settings\ScreenWidth
ActualHeight = XML_Settings\ScreenHeight

Repeat
  
  MBR.l = 0
  MBL.l = 0
  
  ; manage events
  If Not IsFullscreen
    ; check only in windowed mode
    Repeat
      Event = WindowEvent()
      
      Select EventWindow()
          
        Case #MainWindow
      
          Select Event
          
 CompilerIf #PB_Compiler_OS = #PB_OS_Windows
            Case #WM_RBUTTONDOWN
              If WindowMouseY(#MainWindow) < XML_Settings\ScreenHeight : MBR = 1 : EndIf
            Case #WM_LBUTTONDOWN
              If WindowMouseY(#MainWindow) < XML_Settings\ScreenHeight : MBL = 1 : EndIf
            Case #WM_MOUSEWHEEL
              WM_wheel = ((EventwParam() >> 16) & $FFFF)
              WM_wheel = WM_wheel / 120
 CompilerEndIf
            Case #PB_Event_Menu
              Select EventMenu()
                Case #MenuOpen
                  XMLFile$ = OpenFileRequester("Open", GetPathPart(ProgramFilename()), "XML (*.xml)|*.xml", 0 )
                  If XMLFile$
                    If OpenSlideXML(XMLFile$)
                      
                      If ActualWidth <> XML_Settings\ScreenWidth Or ActualHeight <> XML_Settings\ScreenHeight
                        ActualWidth = XML_Settings\ScreenWidth
                        ActualHeight = XML_Settings\ScreenHeight
                        
                        If XML_Settings\ScreenFull = 0
                          ScreenRet = OpenDisplayWindow(); windowed
                        Else
                          ScreenRet = OpenDisplayScreen(); fullscreen-mode
                        EndIf
                        
                      EndIf
                      
                      ; load first page
                      PageCounter = 1
                      If GotoXMLPage(PageCounter)
                        PageStartTime = ElapsedMilliseconds()
                      Else
                        Debug "No valid page information found in XML definition file!"
                        End
                      EndIf
                    EndIf
                  EndIf
                Case #MenuExit
                  Quit = #True
                Case #MenuGoto
                  If OpenWindow(#GotoWindow, 0, 0, 100, 40, "Go to page", #PB_Window_SystemMenu|#PB_Window_WindowCentered, #MainWindow)
                    StringGadget(#PageGadget, 10, 10, 50, 20, "")
                    ButtonGadget(#GoGadget, 70, 10, 30, 20, "Go")
                    SetActiveGadget(#PageGadget)
                    AddKeyboardShortcut(#GotoWindow, #PB_Shortcut_Return, #GoGadget)
                  EndIf
                Case #MenuAbout
                  About$ = "        Pres3D from Kukulkan" + #LF$
                  About$ + #LF$
                  About$ + "kindly given to the PB community" + #LF$
                  About$ + #LF$
                  About$ + "Extended by:" + #LF$
                  About$ + #LF$
                  About$ + "Vera" + #LF$
                  About$ + "Infratec" + #LF$
                  MessageRequester("Pres3D V" + #Version, About$)
              EndSelect
            Case #PB_Event_CloseWindow
              Quit = #True
          EndSelect
          
        Case #GotoWindow
          Select Event
            Case #PB_Event_Gadget, #PB_Event_Menu
              If EventGadget() = #GoGadget
                Help = Val(GetGadgetText(#PageGadget))
                If Help > 0 And Help <= ArraySize(PageNo())
                  ClearObjects()
                  If GotoXMLPage(Help)
                    PageCounter = Help
                    PageStartTime = ElapsedMilliseconds()
                    RemoveKeyboardShortcut(#GotoWindow, #PB_Shortcut_Return)
                    CloseWindow(#GotoWindow)
                  EndIf
                EndIf
              EndIf
            Case #PB_Event_CloseWindow
              RemoveKeyboardShortcut(#GotoWindow, #PB_Shortcut_Return)
              CloseWindow(#GotoWindow)
          EndSelect
      EndSelect
    Until Event = 0
  EndIf
  
  
  ; draw
  ClearScreen(XML_Settings\BackgroundColor)
  ExamineKeyboard()
  ExamineMouse()
  ExamineKeyboard()
  
  ; manage mouse  MouseWheel() works only with a screen so we need also WM_Wheel
  NewZoom = Zoom + MouseWheel() * 20 + WM_Wheel * 20
  WM_Wheel = 0
  ;Debug "current zoom factor: " + Str(Zoom)
  
  If KeyboardPushed(#PB_Key_Return)
    ; clear keyboard input buffer until key is released
    While KeyboardPushed(#PB_Key_Return) : ExamineKeyboard() : Wend
    
    If IsFullscreen : OpenDisplayWindow()
    Else : OpenDisplayScreen()
    EndIf
    
    ; reload current xml file    THIS IS NOT NECCESSARY
    ; OpenSlideXML(XML_CurrentXMLFile.s)
    ClearObjects()  ; is enough
    ; PageCounter = 1
    If GotoXMLPage(PageCounter)
      PageStartTime = ElapsedMilliseconds()
    Else
      Debug "Error reloading XML file " + XML_CurrentXMLFile.s
    EndIf
  EndIf
   
;   If Not Start3D() 
;     Debug "error initializing 3d"
;     End
;   EndIf
  
CompilerIf #PB_Compiler_OS = #PB_OS_Windows
  !extrn _PB_Screen_Direct3DDevice
  !MOV dword EAX, [_PB_Screen_Direct3DDevice]
  !MOV dword [v_pd3d],EAX
  pd3d\SetRenderState(22,1)
CompilerEndIf
  
  ; draw all ingoing effects
  TransformOngoing = #False; used to check, if any animation is not finished (no page-jump possible until this is #False)
  For x = 1 To InObjectCount
    With InObjects(x)
      If ElapsedMilliseconds() - PageStartTime > \Delay 
        If \Transform = #True
          *In.structEffectParameters = \EffectIn
          EffectTransformIn(*In, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, x, InObjectCount)
          TransformOngoing = #True ; some objects are still in progress
          ; transform object with new values
          SetSuperSpritePosition(\SpriteID, \ActualPosX, \ActualPosY, \ActualPosZ + Zoom)
          ;RotateSuperSprite3D(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
          RotateSuperSprite(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
        EndIf
        If \ActualPosZ + Zoom > 0
          If \Type = 0 And XML_Settings\Smoothing = #True
            ;Sprite3DQuality(#PB_Sprite3D_BilinearFiltering) ; text
            SpriteQuality(#PB_Sprite_BilinearFiltering) ; text
          Else
            ;Sprite3DQuality(#PB_Sprite3D_NoFiltering) ; image
            SpriteQuality(#PB_Sprite_NoFiltering) ; image
          EndIf
          ; draw object
          
          If NewZoom <> Zoom
            SetSuperSpritePosition(\SpriteID, \ActualPosX, \ActualPosY, \ActualPosZ + Zoom)
            ;RotateSuperSprite3D(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
            RotateSuperSprite(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
          EndIf
          
          ;DisplaySuperSprite3D(\SpriteID, \ActualTransparency)
          DisplaySuperSprite(\SpriteID, \ActualTransparency)
        EndIf
      Else
        TransformOngoing = #True ; some objectas are still waiting!
      EndIf
    EndWith
  Next  x
  
  ; draw all outgoing effects
  For x = 1 To OutObjectCount
    With OutObjects(x)
      If \Transform = #True        
        TransformOngoing = #True ; some objects are still in progress
        If ElapsedMilliseconds() - PageStartTime > \Delay 
          *Out.structEffectParameters = \EffectOut
          EffectTransformOut(*Out, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, x, OutObjectCount)
          ; transform object with new values
          SetSuperSpritePosition(\SpriteID, \ActualPosX, \ActualPosY, \ActualPosZ + Zoom)
          ;RotateSuperSprite3D(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
          RotateSuperSprite(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
        EndIf
      EndIf
      If \ActualPosZ + Zoom > 0
        If \Type = 0 And XML_Settings\Smoothing = #True
          ;Sprite3DQuality(#PB_Sprite3D_BilinearFiltering) ; text
          SpriteQuality(#PB_Sprite_BilinearFiltering) ; text
        Else
          ;Sprite3DQuality(#PB_Sprite3D_NoFiltering) ; image
          SpriteQuality(#PB_Sprite_NoFiltering) ; image
        EndIf
        ; draw object
        If NewZoom <> Zoom
          SetSuperSpritePosition(\SpriteID, \ActualPosX, \ActualPosY, \ActualPosZ + Zoom)
          ;RotateSuperSprite3D(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
          RotateSuperSprite(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
        EndIf
        ;DisplaySuperSprite3D(\SpriteID, \ActualTransparency)
        DisplaySuperSprite(\SpriteID, \ActualTransparency)
      EndIf
    EndWith
  Next x
  
  Zoom = NewZoom
  
  If Not TransformOngoing
        
    If IsFullscreen
      MBL = MouseButton(#PB_MouseButton_Left)  ; forward
      MBR = MouseButton(#PB_MouseButton_Right) ; back
    EndIf
    
    If KeyboardPushed(#PB_Key_Left)     : MBR = 1: EndIf ; back
    If KeyboardPushed(#PB_Key_Right)    : MBL = 1: EndIf ; forward
    If KeyboardPushed(#PB_Key_PageUp)   : MBR = 1: EndIf ; back
    If KeyboardPushed(#PB_Key_PageDown) : MBL = 1: EndIf ; forward
    If KeyboardPushed(#PB_Key_Space)    : MBL = 1: EndIf ; forward
    
    
    ; page management
    If (MBL And PageCounter <= ArraySize(PageNo())) Or (MBR And PageCounter > 1)
      ; ready with current transforming
      LoadSuccess = #False
      If MBL
        ; left mousebutton pressed (forward)
        PageCounter + 1; initialize next page
        
        ActivateOutgoing() ; activate outgoing effects of current elements
        ; load new page and start ingoing effects
        If LastPageReached = #False
          LoadSuccess = GotoXMLPage(PageCounter)
          PageStartTime = ElapsedMilliseconds()
        EndIf
      Else
        ; right mousebutton pressed (backward)
        PageCounter - 1; initialize previous page
        ActivateOutgoing() ; activate outgoing effects of current elements
        ; load new page and start ingoing effects
        If Not LastPageReached
          LoadSuccess = GotoXMLPage(PageCounter)
        EndIf
        PageStartTime = ElapsedMilliseconds()
      EndIf
  
      If LoadSuccess
        ; successfully initialized new page
      Else
        ; no success with new page
        If LastPageReached = #False
          ; Debug "No more page information found!"
          LastPageReached = #True
        Else
          Quit = #True
        EndIf
        
      EndIf
    EndIf
    If IsStatusBar(0)
      StatusBarText(0, 1, "")
      PageActive = #False
    EndIf
  Else
    If IsStatusBar(0)
      If Not PageActive
        StatusBarText(0, 1, "Active", #PB_StatusBar_Center)
        PageActive = #True
      EndIf
    EndIf
  EndIf
  
  ;Stop3D()
  
  If LastPageReached
    ; show "last page" information
    StartDrawing(ScreenOutput())
    DrawText(1,1, "finished", RGB(100,110,128), RGB(255,255,255))
    StopDrawing()
  EndIf
  
  FlipBuffers()
  Delay(1) ; free some system ressources (keep CPU load below a critical value)

Until Quit Or KeyboardPushed(#PB_Key_Escape)

FreeXML(#XML)

; IDE Options = PureBasic 5.62 (Windows - x86)
; CursorPosition = 322
; FirstLine = 293
; Folding = nP6
; EnableAsm
; EnableXP
; Executable = Pres3D.exe
; CompileSourceDirectory